{-# OPTIONS -fglasgow-exts #-}

--
-- Benchmark tool.
-- Compare a function against equivalent code from other libraries for
-- space and time.
--

import qualified Data.ByteString.Char8 as P
import Data.ByteString (ByteString)

import qualified Data.ByteString as B

import qualified Data.ByteString.ASCII  as A
import qualified Data.ByteString.Latin1 as L
import qualified Data.ByteString.Win1252 as W
import qualified Data.ByteString.UTF8 as U

import Data.List
import Data.Char
import Data.Word

import System.Mem
import Control.Concurrent

import System.IO
import System.CPUTime
import System.IO.Unsafe
import Control.Monad
import Control.Exception
import Text.Printf

main :: IO ()
main = do
    -- initialise
    printf "Initialising test data...\n"
    force (fps,fps') -- >> force (list)

    printf "Size of test data: %dk\n" ((floor $ (fromIntegral (B.length fps)) / 1024) :: Int)
    printf "                Char8\tASCII\tLatin1\tWin1252\tUTF8\tByte\n"

    -- now get to it
    sequence_ . map doit $ tests

doit (s,ls) = do
    printf "%-16s" s
    mapM_ (\s -> time s >> performGC >> threadDelay 100) ls

--  (\s -> time s >> performGC) (head ls) -- only test fps now, 
                                          -- we don't care about beating the other code.
    putChar '\n'
    hFlush stdout

time :: F -> IO ()
time (F a) = do
    start <- getCPUTime
    v     <- force a
    case v of
        B -> printf "--\t"
        _ -> do
            end   <- getCPUTime
            let diff = (fromIntegral (end - start)) / (10^12)
            printf "%0.3f\t" (diff :: Double)
    hFlush stdout

-- compare the 3 implementations
tests  :: [(String,[F])]
tests =
    [
      ("++",    [F ({-# SCC "append" #-}P.append fps fps')
                ,F (A.append fps fps')
                ,F (L.append fps fps'), F (W.append fps fps')
                ,F (U.append fps fps')])

--              ,F () -- SPS.append sps sps)
--              ,F () -- PS.appendPS ps ps)
--              ,F (list ++ list)])

    , ("concat",[F ({-# SCC "concat"    #-}P.concat [fps,fps'])
                ,F (A.concat [fps,fps']), F (L.concat [fps,fps']), F (W.concat [fps,fps'])
                ,F (U.concat [fps,fps'])])

--                 ,F (SPS.concat [sps,sps'])
--                 ,F (PS.concatPS [ps,ps'])
--                 ,F (concat [list,list'])])

    , ("length",[F ({-# SCC "length"    #-}P.length fps)
                ,F (A.length fps), F (L.length fps), F (W.length fps)
                ,F (U.length fps)])

--              ,F (SPS.length sps)
--              ,F (PS.lengthPS ps)
--              ,F (length list)])

{-
    , ("pack",  [F ({-# SCC "pack"      #-}P.pack list)])
--              ,F (SPS.pack list)
--              ,F (PS.packString list)
--              ,F ()])

    , ("unpack",[F ({-# SCC "unpack"    #-}P.unpack fps)])
--              ,F (SPS.unpack sps)
--              ,F (PS.unpackPS ps) ,F ()])
-}

    , ("compare",[F ({-# SCC "compare"   #-}compare fps fps')])

--               ,F (compare sps sps')
--               ,F (compare ps ps')
--               ,F (compare list list')])

    , ("index", [F ({-# SCC "index" #-}P.index fps 10000)
                ,F (A.index fps 10000), F (L.index fps 10000), F (W.index fps 10000)
                ,F () -- (U.index fps 10000)
                ,F (B.index fps 10000)])

--              ,F (SPS.index sps 10000)
--              ,F (PS.indexPS ps 10000)
--              ,F (list !! 10000)])

    , ("map", [F ({-# SCC "map"       #-}P.map id fps)
                ,F (A.map id fps), F (L.map id fps), F (W.map id fps)
              ,F (U.map id fps) ,F (B.map id fps)])

--            ,F (SPS.map toUpper sps)
--            ,F (PS.mapPS toUpper ps)
--            ,F (map toUpper list)])

    , ("filter", [F ({-# SCC "filter"    #-}P.filter (/='f') fps)
                 ,F (A.filter (/='f') fps), F (L.filter (/='f') fps)
                 ,F (W.filter (/='f') fps)
                 ,F (U.filter (/='f') fps)
                 ,F (B.filter (/=102) fps)])

--               ,F (SPS.filter (/='f') sps)
--               ,F (PS.filterPS (/='f') ps)
--               ,F (filter (/='f') list)])

    , ("filterNotChar", [F ({-# SCC "filterNotChar" #-}P.filterNotChar 'f' fps)
                        ,F (A.filterNotChar 'f' fps)
                        , F (L.filterNotChar 'f' fps), F (W.filterNotChar 'f' fps)
                        ,F (U.filterNotChar 'f' fps)
                        ,F (B.filterNotByte 103 fps)])

    , ("findIndex",[F ({-# SCC "findIndex" #-}P.findIndex (=='z') fps)
                   ,F (A.findIndex (=='z') fps), F (L.findIndex (=='z') fps), F (W.findIndex (=='z') fps)
                   ,F (U.findIndex (=='z') fps)
                 ,F (B.findIndex (==122) fps)])

    , ("find",[F ({-# SCC "find"      #-}P.find (=='~') fps)
              ,F (A.find (=='~') fps), F (L.find (=='~') fps)
              ,F (W.find (=='~') fps)
                 ,F (U.find (=='~') fps)
                 ,F (B.find (==122) fps)])
--            ,F (find (=='~') list)])

    , ("foldl", [F ({-# SCC "fold" #-} P.foldl (\a w -> a+1::Int) 0 fps)])

    , ("filterChar", [F ({-# SCC "filterChar"    #-}P.filterChar 'f' fps)
                     ,F (A.filterChar 'f' fps), F (L.filterChar 'f' fps), F (W.filterChar 'f' fps)
                     ,F () -- ,F (U.filterChar 'f' fps)
                     ,F (B.filterByte 103 fps)
                     ])

    , ("take",[F ({-# SCC "take"      #-}P.take 100000 fps)
              ,F (A.take 100000 fps), F (L.take 100000 fps), F (W.take 100000 fps)
              ,F (U.take 100000 fps)
              ])

--           ,F (SPS.take 100000 sps)
--           ,F (PS.takePS 100000 ps)
--           ,F (take 100000 list)])

    , ("drop",[F ({-# SCC "drop"      #-}P.drop 100000 fps)
              ,F (A.drop 100000 fps), F (L.drop 100000 fps), F (W.drop 100000 fps)
              ,F (U.drop 100000 fps)
              ])

--           ,F (SPS.drop 100000 sps)
--           ,F (PS.dropPS 100000 ps)
--           ,F (drop 100000 list)])

    , ("takeWhile",[F ({-# SCC "takeWhile" #-}P.takeWhile (/='z') fps)
                   ,F (A.takeWhile (/='z') fps), F (L.takeWhile (/='z') fps)
                   ,F (W.takeWhile (/='z') fps)
                   ,F (U.takeWhile (/='z') fps)
                   ,F (B.takeWhile (==122) fps)])

--                 ,F (SPS.takeWhile (/='z') sps)
--                 ,F (PS.takeWhilePS (/='z') ps)
--                 ,F (takeWhile (/='z') list)])

    , ("dropWhile",[F ({-# SCC "dropWhile" #-}P.dropWhile (/='z') fps)
                   ,F (A.dropWhile (/='z') fps), F (L.dropWhile (/='z') fps)
                   ,F (W.dropWhile (/='z') fps)
                   ,F (U.dropWhile (/='z') fps)
                   ,F (B.dropWhile (==122) fps)])

--                 ,F (SPS.dropWhile (/='z') sps)
--                 ,F (PS.dropWhilePS (/='z') ps)
--                 ,F (dropWhile (/='z') list)])

    , ("span",[F ({-# SCC "span"      #-}P.span (/='z') fps)
              ,F (A.span (/='z') fps), F (L.span (/='z') fps)
              ,F (W.span (/='z') fps)
              ,F () -- ,F (U.span (/='z') fps)
              ,F (B.span (/=122) fps)
              ])

    , ("spanEnd",[F ({-# SCC "spanEnd"      #-}P.spanEnd (/='z') fps)
                 ,F (A.spanEnd (/='z') fps), F (L.spanEnd (/='z') fps), F (W.spanEnd (/='z') fps)
                 ,F () -- ,F (U.spanEnd (/='z') fps)
                 ,F (B.spanEnd (/=122) fps)])

--                 ,F (SPS.span (/='z') sps)
--                 ,F (PS.spanPS (/='z') ps)
--                 ,F (span (/='z') list)])

    , ("break",[F ({-# SCC "break"     #-}P.break (=='z') fps)
               ,F (A.break (=='z') fps), F (L.break (=='z') fps), F (W.break (=='z') fps)
               ,F () -- ,F (U.break (=='z') fps)
               ,F (B.break (==122) fps)])

--             ,F (SPS.break (=='z') sps)
--             ,F (PS.breakPS (=='z') ps)
--             ,F (break (=='z') list)])

    , ("lines",[F ({-# SCC "lines"     #-}P.lines fps)
               ,F (A.lines fps), F (L.lines fps), F (W.lines fps)
               ,F () -- ,F (U.lines fps)
               ])

--             ,F (SPS.lines sps)
--             ,F (PS.linesPS ps)
--             ,F (lines list)])

    , ("split",[F ({-# SCC "split"     #-}P.split '\n' fps)
               ,F () -- ,F (U.split '\n' fps)
               ,F (B.split 0x0a fps)])

    , ("unlines",[F ({-# SCC "unlines"   #-}P.unlines [fps,fps',fps])
                 ,F (A.unlines [fps,fps',fps]), F (L.unlines [fps,fps',fps]), F (W.unlines [fps,fps',fps])
                 ,F (U.unlines [fps,fps',fps])])
--                 ,F () {-F (SPS.unlines [sps,sps',sps])-}
--                 ,F () -- (PS.unlinesPS [ps,ps',ps])
--                 ,F (unlines [list,list',list])])

    , ("words",[F ({-# SCC "words"     #-}P.words fps)
               ,F (A.words fps), F (L.words fps), F (W.words fps)
               ,F () -- ,F (U.words fps)
               ])
--                 ,F (SPS.words sps)
--                 ,F (PS.wordsPS ps)
--                 ,F (words list)])

    , ("unwords",[F ({-# SCC "unwords"   #-}P.unwords [fps,fps',fps])
                 ,F (A.unwords [fps,fps',fps]), F (L.unwords [fps,fps',fps]), F (W.unwords [fps,fps',fps])
                 ,F (U.unwords [fps,fps',fps]) ])
--                 ,F () {-(SPS.unwords [sps,sps',sps])-}
--                 ,F () -- PS.unwordsPS [ps,ps',ps])
--                 ,F (unwords [list,list',list])])
    , ("group",[F ({-# SCC "group"   #-}P.group fps)
               ,F (A.group fps), F (L.group fps), F (W.group fps)
               ,F (U.group fps)])

    , ("groupBy",[F ({-# SCC "groupBy"   #-}P.groupBy (==) fps)
                 ,F (A.groupBy (==) fps), F (L.groupBy (==) fps), F (W.groupBy (==) fps)
                 ,F (U.groupBy (==) fps)])

    , ("reverse",[F ({-# SCC "reverse"   #-}P.reverse fps)
                 ,F (A.reverse fps), F (L.reverse fps), F (W.reverse fps)
                 ,F (U.reverse fps) ])
--                 ,F (SPS.reverse sps)
--                 ,F (PS.reversePS ps)
--                 ,F (reverse list)])

    , ("cons",[F ({-# SCC "cons"      #-}P.cons 'x' fps)
              ,F (A.cons 'x' fps), F (L.cons 'x' fps), F (W.cons 'x' fps)
              ,F (U.cons 'x' fps)
                 ,F (B.cons 120 fps)])
--            ,F (SPS.cons 'x' sps)
--            ,F (PS.consPS 'x' ps)
--            ,F ('x' : list)])

    , ("snoc",[F ({-# SCC "snoc"      #-}P.snoc fps 'x')
              ,F (A.snoc fps 'x'), F (L.snoc fps 'x'), F (W.snoc fps 'x')
              ,F (U.snoc fps 'x')
                 ,F (B.snoc fps 120)])

    , ("empty",[F ({-# SCC "empty"     #-}P.empty)
               ,F (A.empty), F (L.empty), F (W.empty)
               ,F (U.empty)
               ])
--            ,F (SPS.nil)
--            ,F (PS.nilPS)
--            ,F []])

    , ("head",[F ({-# SCC "head"      #-}P.head fps)
              ,F (A.head fps), F (L.head fps), F (W.head fps)
              ,F (U.head fps)
                 ,F (B.head fps)])
--            ,F (SPS.head sps)
--            ,F (PS.headPS ps)
--            ,F (head list)])

    , ("tail",[F ({-# SCC "tail"      #-}P.tail fps)
              ,F (A.tail fps), F (L.tail fps), F (W.tail fps)
              ,F (U.tail fps) ])
--            ,F (SPS.tail sps)
--            ,F (PS.tailPS ps)
--            ,F (tail list)])

    , ("last",[F ({-# SCC "last"      #-}P.last fps)
              ,F (A.last fps), F (L.last fps), F (W.last fps)
              ,F (U.last fps)
                 ,F (B.last fps)])
--            ,F (last list)])

    , ("init",[F ({-# SCC "init"      #-}P.init fps)
              ,F (A.init fps), F (L.init fps), F (W.init fps)
              ,F (U.init fps)])
--            ,F (init list)])

    , ("inits",[F ({-# SCC "inits"     #-}P.inits fps)
               ,F (A.inits fps), F (L.inits fps), F (W.inits fps)
               ,F (U.inits fps)
               ])

    , ("tails",[F ({-# SCC "tails"     #-}P.tails fps)
               ,F (A.tails fps), F (L.tails fps), F (W.tails fps)
               ,F (U.tails fps)
               ])
--               ,F (tails list)])

    , ("intersperse",[F ({-# SCC "intersperse" #-}P.intersperse 'x' fps)
                     ,F (A.intersperse 'x' fps), F (L.intersperse 'x' fps), F (W.intersperse 'x' fps)
                     ,F (U.intersperse 'x' fps)
                 ,F (B.intersperse 120 fps)])
--            ,F (intersperse 'x' list)])

--  , ("transpose",[F ({-# SCC "transpose" #-}P.transpose [fps,fps'])])

    , ("join",[F ({-# SCC "join" #-}P.join (P.pack "xxx") [fps,fps'])
              ,F (A.join (A.pack "xxx") [fps,fps'])
              ,F (L.join (L.pack "xxx") [fps,fps'])
              ,F (W.join (W.pack "xxx") [fps,fps'])
                 ,F (U.join (U.pack "xxx") [fps,fps'])])

    , ("concatMap",[{-# SCC "concatMap" #-}F ()])
--                ,F (concatMap (\c -> [c]) list)])

    , ("any",[F ({-# SCC "any"       #-}P.any (=='z') fps)
             ,F (A.any (=='z') fps), F (L.any (=='z') fps)
             ,F (W.any (=='z') fps)
                ,F (U.any (=='z') fps)
             ,F (B.any (==120) fps)])
--            ,F (any (=='z') list)])

    , ("all",[F ({-# SCC "all"       #-}P.all (=='z') fps)
             ,F (A.all (=='z') fps), F (L.all (=='z') fps)
             ,F (W.all (=='z') fps)])

    , ("sort",[F ({-# SCC "sort"      #-}P.sort fps)
              ,F (A.sort fps), F (L.sort fps), F (W.sort fps)
              ,F (U.sort fps)])

    , ("maximum",[F ({-# SCC "maximum"   #-}P.maximum fps)
                 ,F (A.maximum fps), F (L.maximum fps), F (W.maximum fps)
                 , F () -- ,F (U.maximum fps)
                 ,F (B.maximum fps)
                 ])
--            ,F (maximum list)])

    , ("minimum",[F ({-# SCC "minimum"   #-}P.minimum fps)
                 ,F (B.minimum fps)
                 ])
--            ,F (minimum list)])


    , ("replicate",[F ({-# SCC "replicate" #-}P.replicate 10000000 'x')
                   ,F (A.replicate 10000000 'x'), F (L.replicate 10000000 'x'), F (W.replicate 10000000 'x')
                   ,F (U.replicate 10000000 'x')])

    , ("splitAt",[F ({-# SCC "splitAt" #-} P.splitAt 10000 fps)
                 ,F (A.splitAt 10000 fps), F (L.splitAt 10000 fps), F (W.splitAt 10000 fps)
                 ,F (U.splitAt 10000 fps)])

    , ("notElem",[F ({-# SCC "notElem"      #-}P.notElem 'z' fps)
                 ,F (A.notElem 'z' fps), F (L.notElem 'z' fps), F (W.notElem 'z' fps)
                 ,F (U.notElem 'z' fps)
                 ,F (B.notElem 122 fps)])

    , ("find",[F ({-# SCC "find"      #-}P.find (=='z') fps)
                 ,F (B.find (==122) fps)])
--            ,F (find (=='z') list)])

    , ("elem",[F ({- SCC "elem"        #-}P.elem 'z' fps)
              ,F (A.elem 'z' fps), F (L.elem 'z' fps), F (W.elem 'z' fps)
              ,F (U.elem 'z' fps)
                 ,F (B.elem 122 fps)])
--            ,F (SPS.elem 'z' sps)
--            ,F (PS.elemPS 'z' ps)
 --           ,F (elem 'z' list)])

    , ("elemIndex",[F ({-# SCC "elemIndex" #-}P.elemIndex 'z' fps)
                   ,F (A.elemIndex 'z' fps), F (L.elemIndex 'z' fps)
                   ,F (W.elemIndex 'z' fps)
                   ,F (U.elemIndex 'z' fps)
                 ,F (B.elemIndex 122 fps)])
--                 ,F (elemIndex 'z' list)])

    , ("elemIndexLast",[F ({-# SCC "elemIndexLast" #-}P.elemIndexLast 'z' fps)
                       ,F (A.elemIndexLast 'z' fps), F (L.elemIndexLast 'z' fps), F (W.elemIndexLast 'z' fps)
                       ,F () -- ,F (U.elemIndexLast 'z' fps)
                 ,F (B.elemIndexLast 122 fps)])

    , ("elemIndices",[F ({-# SCC "elemIndicies" #-} P.elemIndices 'z' fps)
                     ,F (A.elemIndices 'z' fps), F (L.elemIndices 'z' fps), F (W.elemIndices 'z' fps)
                     ,F (U.elemIndices 'z' fps)
                 ,F (B.elemIndices 122 fps)])
--                    ,F (elemIndices 'z' list)])

    , ("findIndices",[F ({-# SCC "findIndicies" #-} P.findIndices (=='z') fps)
                     ,F (A.findIndices (=='z') fps), F (L.findIndices (=='z') fps)
                     ,F (W.findIndices (=='z') fps)
                        ,F (U.findIndices (=='z') fps)
                 ,F (B.findIndices (==122) fps)])

------------------------------------------------------------------------

    , ("lineIndices",[F ({-# SCC "lineIndicies" #-} P.lineIndices fps)
                     ,F (A.lineIndices fps), F (L.lineIndices fps), F (W.lineIndices fps)
                     , F () -- ,F (U.lineIndices fps)
                     ])

    , ("breakChar",[F ({-# SCC "breakChar" #-} P.breakChar 'z' fps)
                   ,F (A.breakChar 'z' fps)
                   , F (L.breakChar 'z' fps), F (W.breakChar 'z' fps)
                   ,F (U.breakChar 'z' fps)
                 ,F (B.breakByte 122 fps)])

    , ("breakSpace",[F ({-# SCC "breakSpace" #-} P.breakSpace fps)
                    ,F (A.breakSpace fps), F (L.breakSpace fps), F (W.breakSpace fps)
                    ,F () -- ,F (U.breakSpace fps)
                    ])

    , ("spanChar",[F ({-# SCC "spanChar" #-} P.spanChar 'z' fps)
                  ,F (A.spanChar 'z' fps), F (L.spanChar 'z' fps), F (W.spanChar 'z' fps)
                  ,F (U.spanChar 'z' fps)
                 ,F (B.spanByte 122 fps)])

    , ("splitWith",[F ({-# SCC "splitWith" #-} P.splitWith (=='z') fps)
                   ,F (A.splitWith (=='z') fps), F (L.splitWith (=='z') fps)
                   ,F (W.splitWith (=='z') fps)
                   ,F () -- ,F (U.splitWith (=='z') fps)
                 ,F (B.splitWith (==122) fps)
                   ])

    , ("dropSpace",[F ({-# SCC "dropSpace" #-} P.dropSpace fps)
                   ,F (A.dropSpace fps), F (L.dropSpace fps), F (W.dropSpace fps)
                   ,F () --, F (U.dropSpace fps)
                   ])

    , ("dropSpaceEnd",[F ({-# SCC "dropSpaceEnd" #-} P.dropSpaceEnd fps)
                      ,F (A.dropSpaceEnd fps), F (L.dropSpaceEnd fps), F (W.dropSpaceEnd fps)
                      ,F () -- ,F (U.dropSpaceEnd fps)
                      ])

    , ("joinWithChar",[F ({-# SCC "joinWithChar" #-} P.joinWithChar ' ' fps fps')
                      ,F (A.joinWithChar ' ' fps fps'), F (L.joinWithChar ' ' fps fps'), F (W.joinWithChar ' ' fps fps')
                      ,F () -- ,F (U.joinWithChar ' ' fps fps')
                 ,F (B.joinWithByte 32 fps fps')])

    , ("join",[F ({-# SCC "join" #-} P.join (P.packChar ' ') [fps,fps'])
              ,F (A.join (A.packChar ' ') [fps,fps']), F (L.join (L.packChar ' ') [fps,fps'])
              ,F (W.join (W.packChar ' ') [fps, fps'])
              ,F (U.join (U.packChar ' ') [fps,fps'])
              ,F (B.join (B.packByte 120) [fps,fps'])])

    , ("zip",[F ({-# SCC "zip" #-} P.zip fps fps)
             ,F (A.zip fps fps), F (L.zip fps fps), F (W.zip fps fps)
             ,F (U.zip fps fps)
             ])
    , ("zipWith",[F ({-# SCC "zipWith" #-} P.zipWith ((. ord) . (+) . ord) fps fps)
                 ,F (A.zipWith ((. ord) . (+) . ord) fps fps), F (L.zipWith ((. ord) . (+) . ord) fps fps)
                 ,F (W.zipWith ((. ord) . (+) . ord) fps fps)
                 ,F (U.zipWith ((. ord) . (+) . ord) fps fps)
                 ])

    , ("isSubstringOf",  [F ({-# SCC "isSubstringOf" #-} P.isSubstringOf (P.pack "email news") fps)
                         ,F (A.isSubstringOf (A.pack "email news") fps), F (L.isSubstringOf (L.pack "email news") fps)
                         , F (W.isSubstringOf (W.pack "email news") fps)
                         , F (U.isSubstringOf (U.pack "email news") fps)])

    , ("isPrefixOf",  [F ({-# SCC "isPrefixOf" #-} P.isPrefixOf (P.pack "The Project Gutenberg eBook") fps)
                      ,F (A.isPrefixOf (A.pack "The Project Gutenberg eBook") fps), F (L.isPrefixOf (L.pack "The Project Gutenberg eBook") fps)
                      ,F (W.isPrefixOf (W.pack "The Project Gutenberg eBook") fps)
                      ,F (U.isPrefixOf (U.pack "The Project Gutenberg eBook") fps)])
    , ("isSuffixOf",  [F ({-# SCC "isSuffixOf" #-} P.isSuffixOf (P.pack "new eBooks") fps)
                      ,F (A.isSuffixOf (A.pack "new eBooks") fps), F (L.isSuffixOf (L.pack "new eBooks") fps)
                      ,F (W.isSuffixOf (W.pack "new eBooks") fps)
                      ,F (U.isSuffixOf (U.pack "new eBooks") fps)])

    , ("count",  [F ({-# SCC "count" #-} P.count '\n' fps)
                 ,F (A.count '\n' fps), F (L.count '\n' fps), F (W.count '\n' fps)
                 ,F (U.count '\n' fps)
                 ])

    , ("addr1",  let s = "my\nstring\nhaskell"# in [F (P.length (P.packAddress s) == 17)
                 ,F (A.length (P.packAddress s) == 17), F (L.length (P.packAddress s) == 17), F (W.length (P.packAddress s) == 17)])
    , ("addr2",  let s = "my\nstring\nhaskell"# in [F (P.unsafePackAddress 17 s == P.packAddress s)
                 ,F (A.unsafePackAddress 17 s == P.packAddress s), F (L.unsafePackAddress 17 s == P.packAddress s), F (W.unsafePackAddress 17 s == P.packAddress s)])

    ]

------------------------------------------------------------------------
-- 
-- an existential list
--
data F = forall a . Forceable a => F a

data Result = T | B

--
-- a bit deepSeqish
--
class Forceable a where
    force :: a -> IO Result
    force v = v `seq` return T

instance Forceable P.ByteString where
    force v = P.length v `seq` return T

-- instance Forceable SPS.PackedString where
--     force v = SPS.length v `seq` return T

-- instance Forceable PS.PackedString where
--     force v = PS.lengthPS v `seq` return T

instance Forceable a => Forceable (Maybe a) where
    force Nothing  = return T
    force (Just v) = force v `seq` return T

instance Forceable [a] where
    force v = length v `seq` return T

instance (Forceable a, Forceable b) => Forceable (a,b) where
    force (a,b) = force a >> force b

instance Forceable Int
instance Forceable Bool
instance Forceable Char
instance Forceable Word8
instance Forceable Ordering

-- used to signal undefinedness
instance Forceable () where force () = return B

------------------------------------------------------------------------
--
-- some large strings to play with
--

fps :: P.ByteString
fps = unsafePerformIO $ P.readFile dict
{-# NOINLINE fps #-}

fps' :: P.ByteString
fps' = unsafePerformIO $ P.readFile dict'
{-# NOINLINE fps' #-}

{-
sps :: SPS.PackedString
sps = unsafePerformIO $ do
    h  <- openFile dict ReadMode
    l  <- hFileSize h
    ps <- SPS.hGet h (fromIntegral l)
    hClose h
    return ps
{-# NOINLINE sps #-}

sps' :: SPS.PackedString
sps' = unsafePerformIO $ do
    h  <- openFile dict' ReadMode
    l  <- hFileSize h
    ps <- SPS.hGet h (fromIntegral l)
    hClose h
    return ps
{-# NOINLINE sps' #-}

ps :: PS.PackedString
ps = unsafePerformIO $ do
    h  <- openFile dict ReadMode
    s  <- hGetContents h
    length s `seq` return ()
    hClose h
    return $! PS.packString s
{-# NOINLINE ps #-}

ps' :: PS.PackedString
ps' = unsafePerformIO $ do
    h  <- openFile dict' ReadMode
    l  <- hFileSize h
    s  <- hGetContents h
    length s `seq` return ()
    hClose h
    return $! PS.packString s
{-# NOINLINE ps' #-}
-}

{-
list :: [Char]
list = unsafePerformIO $ do
    h  <- openFile dict ReadMode
    s  <- hGetContents h
    length s `seq` return ()
    hClose h
    return s
{-# NOINLINE list #-}

list' :: [Char]
list' = unsafePerformIO $ do
    h  <- openFile dict' ReadMode
    s  <- hGetContents h
    length s `seq` return ()
    hClose h
    return s
{-# NOINLINE list' #-}
-}

dict = "bigdata"
dict' = "data"

------------------------------------------------------------------------
